#include <windows.h>
#include <mmsystem.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>

struct mssWaveFormat
{
	DWORD dwFmt;			// must be 0x11 to identify IMA ADPCM
	BYTE  *pData;			// compressed/uncompressed sound data from the file
	DWORD dwDataSize;		// compressed data size
	DWORD dwSampleRate;		// sampling rate (44100/22050/11025)
	DWORD dwBitsPerSample;	// bits per sample (16 - uncompressed PCM / 4 - IMA ADPCM)
	DWORD dwChannels;		// channels (1 - mono / 2 - stereo)
	DWORD dwApproxDataSize;	// approximate resulting data size
	DWORD dwBlockSize;		// ADPCM block size = 0x24 * dwChannels
}mssWavFmt;

FILE *f1;
FILE *f2;
HMODULE hLib;
WAVEFORMATEX wfx;
char wav_file[MAX_PATH];
char snd_file[MAX_PATH];

int (__stdcall *CompressADPCM)(mssWaveFormat *, DWORD **, DWORD **);
int (__stdcall *DecompressADPCM)(mssWaveFormat *, DWORD **, DWORD **);
void (__stdcall *FreeMem)(void *p_mem);
//void (__stdcall *Startup)();
//void (__stdcall *Shutdown)();

void GlobalCleanup()
{
	if ( hLib )
		FreeLibrary(hLib);
	if ( mssWavFmt.pData )
		free(mssWavFmt.pData);
	fcloseall();
//	Shutdown();
}

void prompt(const char *fmt = NULL, ...)
{
	if ( fmt )
	{
		va_list ap;
		va_start(ap, fmt);
		vprintf(fmt, ap);
		va_end(ap);
	}
	else
		printf("\n");

	printf("Press Enter to exit...");
	fflush(stdin);
	getchar();
	exit(1);
}

void InitFunctionPointers()
{
	hLib = LoadLibrary("mss32.dll");
	if ( !hLib )
		prompt("Error: File 'mss32.dll' cannot be loaded!\n");
	CompressADPCM = (int (__stdcall *)(mssWaveFormat *, DWORD **, DWORD **))
		GetProcAddress(hLib, "_AIL_compress_ADPCM@12");
	DecompressADPCM = (int (__stdcall *)(mssWaveFormat *, DWORD **, DWORD **))
		GetProcAddress(hLib, "_AIL_decompress_ADPCM@12");
	FreeMem = (void (__stdcall *)(void *))
		GetProcAddress(hLib, "_AIL_mem_free_lock@4");
/*
	Startup = (void (__stdcall *)())
		GetProcAddress(hLib, "_AIL_startup@0");
	Shutdown = (void (__stdcall *)())
		GetProcAddress(hLib, "_AIL_shutdown@0");
*/
	if ( !CompressADPCM || !DecompressADPCM || !FreeMem )
		prompt("Error: Function was not found in library 'mss32.dll'!\n");
}

DWORD _fsize(FILE *f)
{
	DWORD dwSize;
	DWORD dwCurPos = ftell(f);
	fseek(f, 0, SEEK_END);
	dwSize = ftell(f);
	fseek(f, dwCurPos, SEEK_SET);

	return dwSize;
}

void Init_WaveHeader()
{
	char hdr_id[16];
	if ( _fsize(f1) < 44 )
		prompt("Error: File %s is corrupted or empty\n", wav_file);
	fread(hdr_id, 16, 1, f1);
	memset(&hdr_id[0]+4, 0x20, 4);
	if ( strncmp(hdr_id, "RIFF    WAVEfmt ", 16) )
		prompt("Error: File %s is an invalid wave file\n", wav_file);
	fseek(f1, 20, SEEK_SET);
	fread(&wfx, sizeof(WAVEFORMATEX), 1, f1);
	if ( wfx.wFormatTag != 1 )
		prompt("Error: File %s is already compressed\n", wav_file);

	mssWavFmt.dwFmt = wfx.wFormatTag;
	mssWavFmt.dwSampleRate = wfx.nSamplesPerSec;
	mssWavFmt.dwBitsPerSample = wfx.wBitsPerSample;
	mssWavFmt.dwChannels = wfx.nChannels;
}

void Init_mssHeader()
{
	char *ext = strrchr(snd_file, '.');
	
	mssWavFmt.dwFmt = 0x11;
	mssWavFmt.dwSampleRate = 44100;
	mssWavFmt.dwBitsPerSample = 4;
	if ( ext && (!stricmp(ext, ".adp") || !stricmp(ext, ".wad")) )
		mssWavFmt.dwChannels = 1;
	else
		mssWavFmt.dwChannels = 2;
	mssWavFmt.dwDataSize = _fsize(f1);
	mssWavFmt.dwApproxDataSize = mssWavFmt.dwDataSize * mssWavFmt.dwChannels * 2;
	mssWavFmt.dwBlockSize = mssWavFmt.dwChannels * 0x24;
}

void CompressSound()
{
	DWORD dwSize;
	DWORD *pdwWav, *pdwWavSize;

	fseek(f1, 40, SEEK_SET);
	fread(&dwSize, sizeof(DWORD), 1, f1);

	mssWavFmt.dwDataSize = dwSize;
	mssWavFmt.pData = (BYTE *)malloc(mssWavFmt.dwDataSize);
	mssWavFmt.dwBlockSize = mssWavFmt.dwChannels * 0x24;
	mssWavFmt.dwApproxDataSize = mssWavFmt.dwDataSize / mssWavFmt.dwChannels / 2;
	fread(mssWavFmt.pData, dwSize, 1, f1);

	printf("Compressing file %s...\n"
		   "Format\t\t: PCM (%d)\n"
		   "Sample rate\t: %d\n"
		   "Channels\t: %d\n",
		   wav_file, mssWavFmt.dwFmt, mssWavFmt.dwSampleRate,
		   mssWavFmt.dwChannels);

	if ( CompressADPCM(&mssWavFmt, &pdwWav, &pdwWavSize) )
	{
		memcpy(&dwSize, pdwWav+14, sizeof(DWORD));
		if ( !fwrite(pdwWav+15, dwSize, 1, f2) )
			prompt("Error: Error writing file %s! Disk full?\n", snd_file);
		else
			printf("\nFile was successfully compressed to %s!\n", snd_file);
		FreeMem(pdwWav);
	}
	else
		prompt("Error: Compression failed!\n");
/*
	DWORD cnt = 0;
	printf("\n");
	mssWavFmt.dwDataSize = mssWavFmt.dwChannels * 0x80;
	mssWavFmt.pData = (BYTE *)malloc(mssWavFmt.dwDataSize);
	mssWavFmt.dwApproxDataSize = mssWavFmt.dwDataSize / mssWavFmt.dwChannels / 2 + mssWavFmt.dwChannels * 4;
	mssWavFmt.dwBlockSize = mssWavFmt.dwApproxDataSize;
	for (DWORD i=0; i < dwSize; i+=mssWavFmt.dwDataSize)
	{
		memset(mssWavFmt.pData, 0, mssWavFmt.dwDataSize);
		fread(mssWavFmt.pData, mssWavFmt.dwDataSize, 1, f1);
		if ( CompressADPCM(&mssWavFmt, &pdwWav, &pdwWavSize) )
		{
			if ( !fwrite(pdwWav+15, mssWavFmt.dwBlockSize, 1, f2) )
				prompt("Error: Error writing file %s! Disk full?\n", snd_file);
			else
			{
				if ( cnt != i/(dwSize/100) )
				{
					cnt = i/(dwSize/100);
					printf("Encoding\t: %3d%%\r", cnt);
				}
			}
			FreeMem(pdwWav);
		}
		else
			prompt("Error: Compression failed!\n");
	}
	printf("Encoding\t: Done!\n");
*/
}

void DecompressSound()
{
	DWORD dwSize, *pdwWav, *pdwWavSize;
	mssWavFmt.pData = (BYTE *)malloc(mssWavFmt.dwDataSize);
	if ( !fread(mssWavFmt.pData, mssWavFmt.dwDataSize, 1, f1) )
		prompt("Error: Read failed for file %s\n", snd_file);
	printf("Decompressing file %s...\n"
		   "Format\t\t: IMA ADPCM (%d)\n"
		   "Sample rate\t: %d\n"
		   "Channels\t: %d\n"
		   "ADPCM block size: 0x%02X\n",
		   snd_file, mssWavFmt.dwFmt, mssWavFmt.dwSampleRate,
		   mssWavFmt.dwChannels, mssWavFmt.dwBlockSize);

	if ( DecompressADPCM(&mssWavFmt, &pdwWav, &pdwWavSize) )
	{
		char zero[0x1000] = { 0 };
		memcpy(&dwSize, pdwWav+1, sizeof(DWORD));
		for (DWORD i=0; i < dwSize; i++)
		{
			if ( !memcmp((BYTE *)pdwWav+i, zero, 0x20) )
			{
				if ( !memcmp((BYTE *)pdwWav+i, zero, 0x1000) )
				{
					*(pdwWav+1) = i;
					*(pdwWav+10) = i-36;
					dwSize = i;
					break;
				}
			}
		}
		if ( !fwrite(pdwWav, dwSize, 1, f2) )
			prompt("Error: Error writing file %s! Disk full?\n", wav_file);
		else
			printf("\nFile was successfully decompressed to %s!\n", wav_file);
		FreeMem(pdwWav);
	}
	else
		prompt("Error: Decompression failed!\n");
}

void main(int argc, char *argv[])
{
//	printf("Build on: %s %s\n", __DATE__, __TIME__);
	printf("Enter The Matrix Sound Manager v1.01 - (c) 2003 CoDe RiPPeR\n\n");
	if ( argc < 3 )
	{
		char *exe_name = strrchr(argv[0], '\\');
		if ( !exe_name )
			exe_name = argv[0];
		else
			exe_name++;
		printf("Use: %s [mode] [sound file]\n", exe_name);
		printf("Mode can be:\n"
			   "  -c - compress uncompressed PCM wave file\n"
			   "  -d - decompress encoded sound file into PCM wave file\n\n"
			   "Tip: To process sound files in group, defined by mask, use FOR\n"
			   "command in command line as shown in the following example:\n"
			   "\tfor %%i in (*.adx) do %s -d %%i\n", exe_name);
		prompt();
		return;
	}

	atexit(&GlobalCleanup);
	InitFunctionPointers();
//	Startup();
	
	if ( !stricmp(argv[1], "-c") )
	{
		strcpy(wav_file, argv[2]);
		char *p = strrchr(wav_file, '.');
		if ( p )
			strncpy(snd_file, wav_file, p-&wav_file[0]);
		else
			strcpy(snd_file, wav_file);

		f1 = fopen(wav_file, "rb");
		if ( !f1 )
			prompt("Error: File %s cannot be opened for reading\n", wav_file);
		Init_WaveHeader();
		if ( wfx.nChannels == 1 )
			strcat(snd_file, ".adp");
		else
			strcat(snd_file, ".adx");

		if ( !stricmp(wav_file, snd_file) )
			strcat(snd_file, "_");
		f2 = fopen(snd_file, "wb");
		if ( !f2 )
			prompt("Error: File %s cannot be opened for writing\n", snd_file);

		CompressSound();
	}
	else if ( !stricmp(argv[1], "-d") )
	{
		strcpy(snd_file, argv[2]);
		char *p = strrchr(snd_file, '.');
		if ( p )
			strncpy(wav_file, snd_file, p-&snd_file[0]);
		else
			strcpy(wav_file, snd_file);
		strcat(wav_file, ".wav");

		if ( !stricmp(wav_file, snd_file) )
			strcat(wav_file, "_");

		f1 = fopen(snd_file, "rb");
		if ( !f1 )
			prompt("Error: File %s cannot be opened for reading\n", snd_file);
		Init_mssHeader();
		f2 = fopen(wav_file, "wb");
		if ( !f2 )
			prompt("Error: File %s cannot be opened for writing\n", wav_file);

		DecompressSound();
	}
	else
		prompt("Error: Illegal parameter %s\n", argv[1]);
}